import asyncio
import functools

from fhempy.lib.kia_hyundai.const import VEHICLE_LOCK_ACTION

from .. import fhem, generic, utils
from .utils import get_implementation_by_region_brand
from .Vehicle import Vehicle


class kia_hyundai(generic.FhemModule):

    CAR_BRANDS = {"kia": 1, "hyundai": 2}
    REGION_CODE = {"eu": 1, "us": 2, "ca": 3}

    def __init__(self, logger):
        super().__init__(logger)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        
        attr_config = {
            "update_interval": {
                "default": 30,
                "format": "int",
                "help": "Change interval in minutes, default is 30.",
            }
        }
        await self.set_attr_config(attr_config)

        set_config = {
            "lock": {},
            "unlock": {},
            "start_climate": {
                "args": ["set_temp", "duration", "defrost", "climate", "heating"],
                "help": "Params: Temperature Duration Defrost(on/off) Climate(on/off) Heating(on/off)",
                "params": {
                    "set_temp": {"default": 21, "format": "int"},
                    "duration": {"default": 5, "format": "int"},
                    "defrost": {"default": False, "format": "bool"},
                    "climate": {"default": True, "format": "bool"},
                    "heating": {"default": False, "format": "bool"},
                },
            },
            "stop_climate": {},
            "start_charge": {},
            "stop_charge": {},
            "charge_limits": {
                "args": ["ac_limit", "dc_limit"],
                "help": "Params AC Limit DC Limit",
                "params": {
                    "ac_limit": {"default": 90, "format": "int"},
                    "dc_limit": {"default": 90, "format": "int"},
                },
            },
            "update_data": {},
        }
        await self.set_set_config(set_config)
        
        if len(args) != 8:
            return (
                "Usage: define my_car fhempy kia_hyundai "
                + "USERNAME PASSWORD PIN CAR_BRAND REGION_CODE"
            )
        self.username = args[3]
        self.password = args[4]
        self.pin = args[5]
        self.car_brand = kia_hyundai.CAR_BRANDS[args[6].lower()]
        self.region_code = kia_hyundai.REGION_CODE[args[7].lower()]
        self.create_async_task(self.connect_to_car())

    async def connect_to_car(self):
        self.token = None
        kh_impl = get_implementation_by_region_brand(
            self.region_code, self.car_brand, self.username, self.password
        )
        while self.token is None:
            try:
                self.token = await utils.run_blocking(functools.partial(kh_impl.login))
                if self.token is not None:
                    self.vehicle = Vehicle(
                        self.token, kh_impl, "km", False, str(self.region_code)
                    )
                    self.create_async_task(self.update_loop())
                    break
                await asyncio.sleep(60)
            except Exception:
                self.logger.exception("Failed to login to server, retry in 120s")
                await asyncio.sleep(120)

    async def update_loop(self):
        while True:
            await self.update_once()
            await asyncio.sleep(self._attr_update_interval * 60)

    async def update_once(self):
        try:
            await self.vehicle.update()
            await self.update_readings()
        except Exception:
            self.logger.exception("Failed to update car data")
            await fhem.readingsSingleUpdate(self.hash, "state", "error", 1)

    async def update_readings(self):
        flat_json = utils.flatten_json(self.vehicle.vehicle_data)
        await fhem.readingsBeginUpdate(self.hash)
        try:
            for name in flat_json:
                await fhem.readingsBulkUpdate(self.hash, name, flat_json[name])
            await fhem.readingsBulkUpdate(self.hash, "state", "online")
        except Exception:
            self.logger.exception("Failed to update readings")
        await fhem.readingsEndUpdate(self.hash, 1)

    # Set functions in format: set_NAMEOFSETFUNCTION(self, hash, params)
    async def set_update_data(self, hash, params):
        self.create_async_task(self.update_once())

    async def set_lock(self, hash, params):
        self.create_async_task(self.vehicle.lock_action(VEHICLE_LOCK_ACTION.LOCK))

    async def set_unlock(self, hash, params):
        self.create_async_task(self.vehicle.lock_action(VEHICLE_LOCK_ACTION.UNLOCK))

    async def set_start_climate(self, hash, params):
        self.create_async_task(
            self.vehicle.start_climate(
                params["set_temp"],
                params["duration"],
                params["defrost"],
                params["climate"],
                params["heating"],
            )
        )

    async def set_stop_climate(self, hash, params):
        self.create_async_task(self.vehicle.stop_climate())

    async def set_start_charge(self, hash, params):
        self.create_async_task(self.vehicle.start_charge())

    async def set_stop_charge(self, hash, params):
        self.create_async_task(self.vehicle.stop_charge())

    async def set_charge_limits(self, hash, params):
        self.create_async_task(
            self.vehicle.set_charge_limits(params["ac_limit"], params["dc_limit"])
        )



import asyncio

from aiohttp import ClientSession
from skodaconnect import Connection
from skodaconnect.vehicle import Vehicle

from .. import fhem, generic

# If you wish to use stored tokens, this is an example on how to format data sent to restore_tokens method:
TOKENS = {
    "technical": {"access_token": "...", "refresh_token": "...", "id_token": "..."},
    "connect": {"access_token": "...", "refresh_token": "...", "id_token": "..."},
    "vwg": {"access_token": "...", "refresh_token": "..."},
    "cabs": None,  # Could be populated with access_token, refresh_token, id_token
    "dcs": None,  # Could be populated with access_token, refresh_token, id_token
}
# Comment out the following line to use stored tokens above, set TOKENS=None to do fresh login
TOKENS = None


class skodaconnect(generic.FhemModule):
    def __init__(self, logger):
        super().__init__(logger)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        
        self.attr_config = {
            "vin": {
                "default": "",
                "help": "VIN of your car you want to connect to.",
            },
            "update_interval": {
                "default": 30,
                "format": "int",
                "help": "Readings update intervall in seconds (default 30s).",
            },
            "update_readings": {
                "default": "always",
                "options": "always,onchange",
                "help": "Update readings only on value change or always (default onchange).",
            },
        }
        await self.set_attr_config(self.attr_config)

        if len(args) != 6:
            return (
                "Usage: define my_skoda fhempy skodaconnect my@account.com "
                " PaSSwOrD SPIN"
            )
        self.username = args[3]
        self.password = args[4]
        self.spin = args[5]
        self.create_async_task(self.start_login())
        await fhem.readingsSingleUpdate(self.hash, "state", "connecting", 1)

    async def start_login(self):
        login_success = False
        async with ClientSession(headers={"Connection": "keep-alive"}) as session:
            connection = Connection(session, self.username, self.password, False)
            if TOKENS is not None:
                print("Attempting restore of tokens")
                if await connection.restore_tokens(TOKENS):
                    print("Token restore succeeded")
                    login_success = True
            if not login_success:
                print("Attempting to login to the Skoda Connect service")
                while await connection.doLogin() is False:
                    await asyncio.sleep(5)

            await connection.get_vehicles()
            await connection.update_all()

            self.connection = connection
            if len(connection.vehicles) > 1 and self._attr_vin == "":
                # there is more than one car
                await fhem.readingsSingleUpdate(
                    self.hash, "state", "please set vin attribute", 1
                )
                return

            self.vehicle: Vehicle = None
            for vehicle in connection.vehicles:
                if self._attr_vin != "" and vehicle.vin == self._attr_vin:
                    self.vehicle = vehicle
                elif self._attr_vin == "":
                    self.vehicle = vehicle

            if self.vehicle is None:
                # no car identified
                await fhem.readingsSingleUpdate(self.hash, "state", "no cars found", 1)
                return

            await self.prepare_set_commands()

            await self.update_readings()

    async def prepare_set_commands(self):
        self.set_config = {
            "timer_1": {"args": ["onoff"], "options": "on,off"},
            "timer_2": {"args": ["onoff"], "options": "on,off"},
            "timer_3": {"args": ["onoff"], "options": "on,off"},
            "timer_1_schedule": {
                "args": ["active", "recurring", "date", "time", "days"],
                "help": (
                    "Parameters: activate(0/1) recurring(0/1) date(yyyymmdd) time(hhmm) weekdays(yyyyynn)<br>"
                    "e.g. set my_skoda timer_1_schedule 1 1 0800 yyyyynn<br>"
                    "recurring timer which is activated mon-fri 08:00<br>"
                    "e.g. set my_skoda timer_1_schedule 1 0 20250101 0800<br>"
                    "timer which is only activated on 1.1.2025 at 08:00"
                ),
            },
            "timer_2_schedule": {
                "args": ["active", "recurring", "date", "time", "days"],
                "help": (
                    "Parameters: activate(0/1) recurring(0/1) date(yyyymmdd) time(hhmm) weekdays(yyyyynn)<br>"
                    "e.g. set my_skoda timer_1_schedule 1 1 0800 yyyyynn<br>"
                    "recurring timer which is activated mon-fri 08:00<br>"
                    "e.g. set my_skoda timer_1_schedule 1 0 20250101 0800<br>"
                    "timer which is only activated on 1.1.2025 at 08:00"
                ),
            },
            "timer_3_schedule": {
                "args": ["active", "recurring", "date", "time", "days"],
                "help": (
                    "Parameters: activate(0/1) recurring(0/1) date(yyyymmdd) time(hhmm) weekdays(yyyyynn)<br>"
                    "e.g. set my_skoda timer_1_schedule 1 1 0800 yyyyynn<br>"
                    "recurring timer which is activated mon-fri 08:00<br>"
                    "e.g. set my_skoda timer_1_schedule 1 0 20250101 0800<br>"
                    "timer which is only activated on 1.1.2025 at 08:00"
                ),
            },
        }
        if self.vehicle.is_charging_supported:
            self.set_config["charger"] = {
                "args": ["onoff"],
                "options": "on,off",
            }
            self.set_config["charger_current"] = {
                "args": ["current"],
                "options": "slider,1,1,254",
            }
            self.set_config["charge_limit"] = {
                "args": ["limit"],
                "options": "0,10,20,30,40,50",
            }
        if self.vehicle.is_electric_climatisation_supported:
            self.set_config["battery_climatisation"] = {
                "args": ["onoff"],
                "options": "on,off",
            }
            self.set_config["climatisation"] = {
                "args": ["mode"],
                "options": "auxiliary,electric,off",
            }
        if (
            self.vehicle.is_auxiliary_climatisation_supported
            or self.vehicle.is_electric_climatisation_supported
        ):
            self.set_config["climatisation_target_temperature"] = {
                "args": ["temperature"],
                "params": {"temperature": {"format": "int"}},
                "options": "slider,16,1,30",
            }
        if self.vehicle.is_window_heater_supported:
            self.set_config["window_heating"] = {
                "args": ["startstop"],
                "options": "start,stop",
            }

        self.set_config["lock"] = {
            "args": ["lockunlock"],
            "options": "lock,unlock",
        }

        self.set_config["honkandflash"] = {
            "args": ["honkandflash"],
            "options": "flash,honkandflash",
        }

        if self.vehicle.is_pheater_heating_supported:
            self.set_config["pheater"] = {
                "args": ["mode"],
                "options": "heating,ventilation,off",
            }
            self.set_config["pheater_duration"] = {
                "args": ["duration"],
                "options": "slider,10,10,60",
                "params": {"duration": {"format": "int"}},
            }

        self.set_config["force_update"] = {
            "help": (
                "will trigger force update<br>"
                "this command should not be used too often<br>"
                "you might be locked out by Skoda<br>"
                "until next motor start"
            ),
        }

        await self.set_set_config(self.set_config)

    async def set_pheater(self, hash, params):
        self.create_async_task(self.vehicle.set_pheater(params["mode"], self.spin))

    async def set_pheater_duration(self, hash, params):
        self.vehicle.pheater_duration = params["duration"]
        self.create_async_task(self.update_readings_once())

    async def set_lock(self, hash, params):
        self.create_async_task(self.vehicle.set_lock(params["lockunlock"], self.spin))

    async def set_honkandflash(self, hash, params):
        self.create_async_task(self.vehicle.set_honkandflash(params["honkandflash"]))

    async def set_charger(self, hash, params):
        self.create_async_task(self.vehicle.set_charger(params["onoff"]))

    async def set_charger_current(self, hash, params):
        self.create_async_task(self.vehicle.set_charger_current(params["current"]))

    async def set_charge_limit(self, hash, params):
        self.create_async_task(self.vehicle.set_charge_limit(params["limit"]))

    async def set_battery_climatisation(self, hash, params):
        self.create_async_task(self.vehicle.set_battery_climatisation(params["onoff"]))

    async def set_climatisation(self, hash, params):
        self.create_async_task(
            self.vehicle.set_climatisation(params["mode"], spin=self.spin)
        )

    async def set_climatisation_target_temperature(self, hash, params):
        self.create_async_task(
            self.vehicle.set_climatisation_temp(params["temperature"])
        )

    async def set_window_heating(self, hash, params):
        self.create_async_task(self.vehicle.set_window_heating(params["startstop"]))

    async def set_force_update(self, hash, params):
        self.create_async_task(self.vehicle.set_refresh())

    async def update_readings(self):
        self.instruments = self.vehicle.dashboard(mutable=True).instruments
        while True:
            await self.update_readings_once()
            await asyncio.sleep(self._attr_update_interval)

    async def update_readings_once(self):
        await self.connection.update_all()
        try:
            for instrument in self.instruments:
                if hasattr(instrument, "reverse_state") and instrument.reverse_state:
                    val_state = not instrument.state
                else:
                    val_state = instrument.state
                if self._attr_update_readings == "always":
                    await fhem.readingsSingleUpdate(
                        self.hash, instrument.attr, val_state, 1
                    )
                    await fhem.readingsSingleUpdate(
                        self.hash, instrument.attr + "_str", instrument.str_state, 1
                    )
                else:
                    await fhem.readingsSingleUpdateIfChanged(
                        self.hash, instrument.attr, val_state, 1
                    )
                    await fhem.readingsSingleUpdateIfChanged(
                        self.hash, instrument.attr + "_str", instrument.str_state, 1
                    )
                if instrument.attr == "fuel_level":
                    if self._attr_update_readings == "always":
                        await fhem.readingsSingleUpdate(
                            self.hash, "state", instrument.str_state, 1
                        )
                    else:
                        await fhem.readingsSingleUpdateIfChanged(
                            self.hash, "state", instrument.str_state, 1
                        )
        except Exception:
            self.logger.exception("Failed to update readings")



import json
import logging
import random
import time
import uuid
from datetime import datetime, timedelta
from urllib.parse import parse_qs, urlparse

import curlify
import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.ssl_ import create_urllib3_context

from .const import (
    BRAND_HYUNDAI,
    BRAND_KIA,
    BRANDS,
    DATE_FORMAT,
    DOMAIN,
    VEHICLE_LOCK_ACTION,
)
from .KiaUvoApiImpl import KiaUvoApiImpl
from .Token import Token

CIPHERS = "DEFAULT@SECLEVEL=1"

_LOGGER = logging.getLogger(__name__)


class cipherAdapter(HTTPAdapter):
    """
    A HTTPAdapter that re-enables poor ciphers required by Hyundai.
    """

    def init_poolmanager(self, *args, **kwargs):
        context = create_urllib3_context(ciphers=CIPHERS)
        kwargs["ssl_context"] = context
        return super().init_poolmanager(*args, **kwargs)

    def proxy_manager_for(self, *args, **kwargs):
        context = create_urllib3_context(ciphers=CIPHERS)
        kwargs["ssl_context"] = context
        return super().proxy_manager_for(*args, **kwargs)


class HyundaiBlueLinkAPIUSA(KiaUvoApiImpl):

    old_vehicle_status = None
    # initialize with a timestamp which will allow the first fetch to occur
    last_loc_timestamp = datetime.now() - timedelta(hours=3)

    def __init__(
        self,
        username: str,
        password: str,
        region: int,
        brand: int,
        use_email_with_geocode_api: bool = False,
        pin: str = "",
    ):
        super().__init__(
            username, password, region, brand, use_email_with_geocode_api, pin
        )

        self.BASE_URL: str = "api.telematics.hyundaiusa.com"
        self.LOGIN_API: str = "https://" + self.BASE_URL + "/v2/ac/"
        self.API_URL: str = "https://" + self.BASE_URL + "/ac/v2/"

        ts = time.time()
        utc_offset = (
            datetime.fromtimestamp(ts) - datetime.utcfromtimestamp(ts)
        ).total_seconds()
        utc_offset_hours = int(utc_offset / 60 / 60)

        self.API_HEADERS = {
            "content-type": "application/json;charset=UTF-8",
            "accept": "application/json, text/plain, */*",
            "accept-encoding": "gzip, deflate, br",
            "accept-language": "en-US,en;q=0.9",
            "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/75.0.3770.142 Safari/537.36",
            "host": self.BASE_URL,
            "origin": "https://" + self.BASE_URL,
            "referer": "https://" + self.BASE_URL + "/login",
            "from": "SPA",
            "to": "ISS",
            "language": "0",
            "offset": str(utc_offset_hours),
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "refresh": "false",
            "encryptFlag": "false",
            "brandIndicator": "H",
            "gen": "2",
            "username": self.username,
            "blueLinkServicePin": self.pin,
            "client_id": "m66129Bb-em93-SPAHYN-bZ91-am4540zp19920",
            "clientSecret": "v558o935-6nne-423i-baa8",
        }
        self.sessions = requests.Session()
        self.sessions.mount("https://" + self.BASE_URL, cipherAdapter())

        _LOGGER.debug(f"{DOMAIN} - initial API headers: {self.API_HEADERS}")

    def login(self) -> Token:
        username = self.username
        password = self.password

        ### Sign In with Email and Password and Get Authorization Code ###

        url = self.LOGIN_API + "oauth/token"

        data = {"username": username, "password": password}
        headers = self.API_HEADERS
        response = self.sessions.post(url, json=data, headers=headers)
        _LOGGER.debug(f"{DOMAIN} - Sign In Response {response.text}")
        response = response.json()
        access_token = response["access_token"]
        refresh_token = response["refresh_token"]
        expires_in = float(response["expires_in"])
        _LOGGER.debug(f"{DOMAIN} - Access Token Value {access_token}")
        _LOGGER.debug(f"{DOMAIN} - Refresh Token Value {refresh_token}")

        ### Get Vehicles ###
        response = self.get_vehicle(access_token)
        vehicle_details = response["enrolledVehicleDetails"][0]["vehicleDetails"]
        vehicle_name = vehicle_details["nickName"]
        vehicle_id = vehicle_details["vin"]
        vehicle_regid = vehicle_details["regid"]
        _LOGGER.debug(f"{DOMAIN} - vehicle_regid={vehicle_regid}")
        vehicle_model = vehicle_details["modelCode"]
        vehicle_registration_date = vehicle_details["enrollmentDate"]

        valid_until = (datetime.now() + timedelta(seconds=expires_in)).strftime(
            DATE_FORMAT
        )

        token = Token({})
        token.set(
            access_token,
            refresh_token,
            None,
            vehicle_name,
            vehicle_id,
            vehicle_regid,
            vehicle_model,
            vehicle_registration_date,
            valid_until,
            "NoStamp",
        )

        _LOGGER.debug(f"{DOMAIN} - updated API headers: {self.API_HEADERS}")

        return token

    def get_cached_vehicle_status(self, token: Token):
        # Vehicle Status Call
        url = self.API_URL + "rcs/rvs/vehicleStatus"
        headers = self.API_HEADERS
        headers["accessToken"] = token.access_token
        headers["vin"] = token.vehicle_id

        _LOGGER.debug(f"{DOMAIN} - using API headers: {self.API_HEADERS}")

        response = self.sessions.get(url, headers=headers)
        response = response.json()
        _LOGGER.debug(f"{DOMAIN} - get_cached_vehicle_status response {response}")

        vehicle_status = {}
        vehicle_status["vehicleStatus"] = response["vehicleStatus"]

        vehicle_status["vehicleStatus"]["dateTime"] = (
            vehicle_status["vehicleStatus"]["dateTime"]
            .replace("-", "")
            .replace("T", "")
            .replace(":", "")
            .replace("Z", "")
        )
        vehicle_status["vehicleStatus"]["time"] = vehicle_status["vehicleStatus"][
            "dateTime"
        ]
        vehicle_status["vehicleStatus"]["date"] = vehicle_status["vehicleStatus"][
            "dateTime"
        ]
        vehicle_status["vehicleStatus"]["doorLock"] = vehicle_status["vehicleStatus"][
            "doorLock"
        ]
        vehicle_status["vehicleLocation"] = vehicle_status["vehicleStatus"][
            "vehicleLocation"
        ]
        if vehicle_status["vehicleStatus"].get("tirePressureLamp"):
            vehicle_status["vehicleStatus"]["tirePressureLamp"][
                "tirePressureLampAll"
            ] = vehicle_status["vehicleStatus"]["tirePressureLamp"][
                "tirePressureWarningLampAll"
            ]
            vehicle_status["vehicleStatus"]["tirePressureLamp"][
                "tirePressureLampFL"
            ] = vehicle_status["vehicleStatus"]["tirePressureLamp"][
                "tirePressureWarningLampFrontLeft"
            ]
            vehicle_status["vehicleStatus"]["tirePressureLamp"][
                "tirePressureLampFR"
            ] = vehicle_status["vehicleStatus"]["tirePressureLamp"][
                "tirePressureWarningLampFrontRight"
            ]
            vehicle_status["vehicleStatus"]["tirePressureLamp"][
                "tirePressureLampRR"
            ] = vehicle_status["vehicleStatus"]["tirePressureLamp"][
                "tirePressureWarningLampRearRight"
            ]
            vehicle_status["vehicleStatus"]["tirePressureLamp"][
                "tirePressureLampRL"
            ] = vehicle_status["vehicleStatus"]["tirePressureLamp"][
                "tirePressureWarningLampRearLeft"
            ]
        # Get Odomoter Details
        response = self.get_vehicle(token.access_token)
        vehicle_status["odometer"] = {}
        vehicle_status["odometer"]["unit"] = 3
        vehicle_status["odometer"]["value"] = response["enrolledVehicleDetails"][0][
            "vehicleDetails"
        ]["odometer"]

        vehicle_status["vehicleLocation"] = self.get_location(
            token, vehicle_status["odometer"]["value"]
        )
        HyundaiBlueLinkAPIUSA.old_vehicle_status = vehicle_status
        return vehicle_status

    def get_location(self, token: Token, current_odometer):
        r"""
        Get the location of the vehicle

        Only update the location if the odometer moved AND if the last location update was over an hour ago.
        Note that the "last updated" time is initially set to three hours ago.

        This will help to prevent too many cals to the API
        """
        url = self.API_URL + "rcs/rfc/findMyCar"
        headers = self.API_HEADERS
        headers["accessToken"] = token.access_token
        headers["vehicleId"] = token.vehicle_id
        headers["pAuth"] = self.get_pin_token(token)
        prev_odometer = "0"
        if HyundaiBlueLinkAPIUSA.old_vehicle_status is not None:
            prev_odometer = HyundaiBlueLinkAPIUSA.old_vehicle_status.get(
                "odometer", {}
            ).get("value")
        check_server = (
            current_odometer > prev_odometer
            and HyundaiBlueLinkAPIUSA.last_loc_timestamp
            < datetime.now() - timedelta(hours=1)
        )
        if check_server:
            try:
                HyundaiBlueLinkAPIUSA.last_loc_timestamp = datetime.now()
                response = self.sessions.get(url, headers=headers)
                response_json = response.json()
                _LOGGER.debug(f"{DOMAIN} - Get Vehicle Location {response_json}")
                if response_json.get("coord") is not None:
                    return response_json
                else:
                    # Check for rate limit exceeded
                    # These hard-coded values were extracted from a rate limit exceeded response.  In either case the log
                    # will include the full response when the "coord" attribute is not present
                    if (
                        response_json.get("errorCode", 0) == 502
                        and response_json.get("errorSubCode", "") == "HT_534"
                    ):
                        # rate limit exceeded; set the last_loc_timestamp such that the next check will be at least 12 hours from now
                        HyundaiBlueLinkAPIUSA.last_loc_timestamp = (
                            datetime.now() + timedelta(hours=11)
                        )
                        _LOGGER.warn(
                            f"{DOMAIN} - get vehicle location rate limit exceeded.  Location will not be fetched until at least {HyundaiBlueLinkAPIUSA.last_loc_timestamp + timedelta(hours = 12)}"
                        )
                    else:
                        _LOGGER.warn(
                            f"{DOMAIN} - Unable to get vehicle location: {response_json}"
                        )

                    if HyundaiBlueLinkAPIUSA.old_vehicle_status is not None:
                        return HyundaiBlueLinkAPIUSA.old_vehicle_status.get(
                            "vehicleLocation"
                        )
                    else:
                        return None

            except Exception as e:
                _LOGGER.warning(
                    f"{DOMAIN} - Get vehicle location failed: {e}", exc_info=True
                )
                if HyundaiBlueLinkAPIUSA.old_vehicle_status is not None:
                    return HyundaiBlueLinkAPIUSA.old_vehicle_status.get(
                        "vehicleLocation"
                    )
                else:
                    return None
        elif HyundaiBlueLinkAPIUSA.old_vehicle_status is not None:
            return HyundaiBlueLinkAPIUSA.old_vehicle_status.get("vehicleLocation")

    def get_vehicle(self, access_token):
        username = self.username
        password = self.password

        url = self.API_URL + "enrollment/details/" + username
        headers = self.API_HEADERS
        headers["accessToken"] = access_token
        response = self.sessions.get(url, headers=headers)
        _LOGGER.debug(f"{DOMAIN} - Get Vehicles Response {response.text}")
        response = response.json()

        return response

    def get_pin_token(self, token: Token):
        pass

    def update_vehicle_status(self, token: Token):
        pass

    def lock_action(self, token: Token, action):
        _LOGGER.debug(f"{DOMAIN} - Action for lock is: {action}")

        if action == "close":
            url = self.API_URL + "rcs/rdo/off"
            _LOGGER.debug(f"{DOMAIN} - Calling Lock")
        else:
            url = self.API_URL + "rcs/rdo/on"
            _LOGGER.debug(f"{DOMAIN} - Calling unlock")

        headers = self.API_HEADERS
        headers["accessToken"] = token.access_token
        headers["vin"] = token.vehicle_id
        headers["registrationId"] = token.vehicle_regid
        headers["APPCLOUD-VIN"] = token.vehicle_id

        data = {"userName": self.username, "vin": token.vehicle_id}
        response = self.sessions.post(url, headers=headers, json=data)
        # response_headers = response.headers
        # response = response.json()
        # action_status = self.check_action_status(token, headers["pAuth"], response_headers["transactionId"])

        # _LOGGER.debug(f"{DOMAIN} - Received lock_action response {action_status}")
        _LOGGER.debug(
            f"{DOMAIN} - Received lock_action response status code: {response.status_code}"
        )
        _LOGGER.debug(f"{DOMAIN} - Received lock_action response: {response.text}")

    def start_climate(
        self, token: Token, set_temp, duration, defrost, climate, heating
    ):
        _LOGGER.debug(f"{DOMAIN} - Start engine..")

        url = self.API_URL + "rcs/rsc/start"

        headers = self.API_HEADERS
        headers["accessToken"] = token.access_token
        headers["vin"] = token.vehicle_id
        headers["registrationId"] = token.vehicle_regid
        _LOGGER.debug(f"{DOMAIN} - Start engine headers: {headers}")

        data = {
            "Ims": 0,
            "airCtrl": int(climate),
            "airTemp": {"unit": 1, "value": set_temp},
            "defrost": defrost,
            "heating1": int(heating),
            "igniOnDuration": duration,
            # "seatHeaterVentInfo": None,
            "username": self.username,
            "vin": token.vehicle_id,
        }
        _LOGGER.debug(f"{DOMAIN} - Start engine data: {data}")

        response = self.sessions.post(url, json=data, headers=headers)

        # _LOGGER.debug(f"{DOMAIN} - Start engine curl: {curlify.to_curl(response.request)}")
        _LOGGER.debug(
            f"{DOMAIN} - Start engine response status code: {response.status_code}"
        )
        _LOGGER.debug(f"{DOMAIN} - Start engine response: {response.text}")

    def stop_climate(self, token: Token):
        _LOGGER.debug(f"{DOMAIN} - Stop engine..")

        url = self.API_URL + "rcs/rsc/stop"

        headers = self.API_HEADERS
        headers["accessToken"] = token.access_token
        headers["vin"] = token.vehicle_id
        headers["registrationId"] = token.vehicle_regid

        _LOGGER.debug(f"{DOMAIN} - Stop engine headers: {headers}")

        response = self.sessions.post(url, headers=headers)
        _LOGGER.debug(
            f"{DOMAIN} - Stop engine response status code: {response.status_code}"
        )
        _LOGGER.debug(f"{DOMAIN} - Stop engine response: {response.text}")

    def start_charge(self, token: Token):
        pass

    def stop_charge(self, token: Token):
        pass



class bodyScales:
    def __init__(self, age, height, gender, weight, scaleType="xiaomi"):
        self.age = age
        self.height = height
        self.gender = gender
        self.weight = weight

        if scaleType == "xiaomi":
            self.scaleType = "xiaomi"
        else:
            self.scaleType = "holtek"

    # Get BMI scale
    def getBMIScale(self):
        if self.scaleType == "xiaomi":
            # Amazfit/new mi fit
            # return [18.5, 24, 28]
            # Old mi fit // amazfit for body figure
            return [18.5, 25.0, 28.0, 32.0]
        elif self.scaleType == "holtek":
            return [18.5, 25.0, 30.0]

    # Get fat percentage scale
    def getFatPercentageScale(self):
        # The included tables where quite strange, maybe bogus, replaced them with better ones...
        if self.scaleType == "xiaomi":
            scales = [
                {
                    "min": 0,
                    "max": 12,
                    "female": [12.0, 21.0, 30.0, 34.0],
                    "male": [7.0, 16.0, 25.0, 30.0],
                },
                {
                    "min": 12,
                    "max": 14,
                    "female": [15.0, 24.0, 33.0, 37.0],
                    "male": [7.0, 16.0, 25.0, 30.0],
                },
                {
                    "min": 14,
                    "max": 16,
                    "female": [18.0, 27.0, 36.0, 40.0],
                    "male": [7.0, 16.0, 25.0, 30.0],
                },
                {
                    "min": 16,
                    "max": 18,
                    "female": [20.0, 28.0, 37.0, 41.0],
                    "male": [7.0, 16.0, 25.0, 30.0],
                },
                {
                    "min": 18,
                    "max": 40,
                    "female": [21.0, 28.0, 35.0, 40.0],
                    "male": [11.0, 17.0, 22.0, 27.0],
                },
                {
                    "min": 40,
                    "max": 60,
                    "female": [22.0, 29.0, 36.0, 41.0],
                    "male": [12.0, 18.0, 23.0, 28.0],
                },
                {
                    "min": 60,
                    "max": 100,
                    "female": [23.0, 30.0, 37.0, 42.0],
                    "male": [14.0, 20.0, 25.0, 30.0],
                },
            ]

        elif self.scaleType == "holtek":
            scales = [
                {
                    "min": 0,
                    "max": 21,
                    "female": [18, 23, 30, 35],
                    "male": [8, 14, 21, 25],
                },
                {
                    "min": 21,
                    "max": 26,
                    "female": [19, 24, 30, 35],
                    "male": [10, 15, 22, 26],
                },
                {
                    "min": 26,
                    "max": 31,
                    "female": [20, 25, 31, 36],
                    "male": [11, 16, 21, 27],
                },
                {
                    "min": 31,
                    "max": 36,
                    "female": [21, 26, 33, 36],
                    "male": [13, 17, 25, 28],
                },
                {
                    "min": 36,
                    "max": 41,
                    "female": [22, 27, 34, 37],
                    "male": [15, 20, 26, 29],
                },
                {
                    "min": 41,
                    "max": 46,
                    "female": [23, 28, 35, 38],
                    "male": [16, 22, 27, 30],
                },
                {
                    "min": 46,
                    "max": 51,
                    "female": [24, 30, 36, 38],
                    "male": [17, 23, 29, 31],
                },
                {
                    "min": 51,
                    "max": 56,
                    "female": [26, 31, 36, 39],
                    "male": [19, 25, 30, 33],
                },
                {
                    "min": 56,
                    "max": 100,
                    "female": [27, 32, 37, 40],
                    "male": [21, 26, 31, 34],
                },
            ]

        for scale in scales:
            if self.age >= scale["min"] and self.age < scale["max"]:
                return scale[self.gender]

    # Get muscle mass scale
    def getMuscleMassScale(self):
        if self.scaleType == "xiaomi":
            scales = [
                {
                    "min": {"male": 170, "female": 160},
                    "female": [36.5, 42.6],
                    "male": [49.4, 59.5],
                },
                {
                    "min": {"male": 160, "female": 150},
                    "female": [32.9, 37.6],
                    "male": [44.0, 52.5],
                },
                {
                    "min": {"male": 0, "female": 0},
                    "female": [29.1, 34.8],
                    "male": [38.5, 46.6],
                },
            ]
        elif self.scaleType == "holtek":
            scales = [
                {
                    "min": {"male": 170, "female": 170},
                    "female": [36.5, 42.5],
                    "male": [49.5, 59.4],
                },
                {
                    "min": {"male": 160, "female": 160},
                    "female": [32.9, 37.5],
                    "male": [44.0, 52.4],
                },
                {
                    "min": {"male": 0, "female": 0},
                    "female": [29.1, 34.7],
                    "male": [38.5, 46.5],
                },
            ]

        for scale in scales:
            if self.height >= scale["min"][self.gender]:
                return scale[self.gender]

    # Get water percentage scale
    def getWaterPercentageScale(self):
        if self.scaleType == "xiaomi":
            if self.gender == "male":
                return [55.0, 65.1]
            elif self.gender == "female":
                return [45.0, 60.1]
        elif self.scaleType == "holtek":
            return [53, 67]

    # Get visceral fat scale
    def getVisceralFatScale(self):
        # Actually the same in mi fit/amazfit and holtek's sdk
        return [10.0, 15.0]

    # Get bone mass scale
    def getBoneMassScale(self):
        if self.scaleType == "xiaomi":
            scales = [
                {
                    "male": {"min": 75.0, "scale": [2.0, 4.2]},
                    "female": {"min": 60.0, "scale": [1.8, 3.9]},
                },
                {
                    "male": {"min": 60.0, "scale": [1.9, 4.1]},
                    "female": {"min": 45.0, "scale": [1.5, 3.8]},
                },
                {
                    "male": {"min": 0.0, "scale": [1.6, 3.9]},
                    "female": {"min": 0.0, "scale": [1.3, 3.6]},
                },
            ]

            for scale in scales:
                if self.weight >= scale[self.gender]["min"]:
                    return scale[self.gender]["scale"]

        elif self.scaleType == "holtek":
            scales = [
                {
                    "female": {"min": 60, "optimal": 2.5},
                    "male": {"min": 75, "optimal": 3.2},
                },
                {
                    "female": {"min": 45, "optimal": 2.2},
                    "male": {"min": 69, "optimal": 2.9},
                },
                {
                    "female": {"min": 0, "optimal": 1.8},
                    "male": {"min": 0, "optimal": 2.5},
                },
            ]

            for scale in scales:
                if self.weight >= scale[self.gender]["min"]:
                    return [
                        scale[self.gender]["optimal"] - 1,
                        scale[self.gender]["optimal"] + 1,
                    ]

    # Get BMR scale
    def getBMRScale(self):
        if self.scaleType == "xiaomi":
            coefficients = {
                "male": {30: 21.6, 50: 20.07, 100: 19.35},
                "female": {30: 21.24, 50: 19.53, 100: 18.63},
            }
        elif self.scaleType == "holtek":
            coefficients = {
                "female": {12: 34, 15: 29, 17: 24, 29: 22, 50: 20, 120: 19},
                "male": {12: 36, 15: 30, 17: 26, 29: 23, 50: 21, 120: 20},
            }

        for age, coefficient in coefficients[self.gender].items():
            if self.age < age:
                return [self.weight * coefficient]

    # Get protein scale (hardcoded in mi fit)
    def getProteinPercentageScale(self):
        # Actually the same in mi fit and holtek's sdk
        return [16, 20]

    # Get ideal weight scale (BMI scale converted to weights)
    def getIdealWeightScale(self):
        scale = []
        for bmiScale in self.getBMIScale():
            scale.append((bmiScale * self.height) * self.height / 10000)
        return scale

    # Get Body Score scale
    def getBodyScoreScale(self):
        # very bad, bad, normal, good, better
        return [50.0, 60.0, 80.0, 90.0]

    # Return body type scale
    def getBodyTypeScale(self):
        return [
            "obese",
            "overweight",
            "thick-set",
            "lack-exercise",
            "balanced",
            "balanced-muscular",
            "skinny",
            "balanced-skinny",
            "skinny-muscular",
        ]



import asyncio
import functools

import aiohttp
from bs4 import BeautifulSoup

from .. import fhem, generic, utils


class geizhals(generic.FhemModule):

    headers = {
        "accept": "application/json",
        "accept-language": "en-AT,en;q=0.9,de-AT;q=0.8,de;q=0.7,en-GB;q=0.6,en-US;q=0.5",
        "content-type": "application/json",
        "sec-ch-ua": '"Chromium";v="104", " Not A;Brand";v="99", "Google Chrome";v="104"',
        "sec-ch-ua-mobile": "?0",
        "sec-ch-ua-platform": '"Chrome OS"',
        "sec-fetch-dest": "empty",
        "sec-fetch-mode": "cors",
        "sec-fetch-site": "same-origin",
        "Referer": "https://geizhals.at/",
        "Referrer-Policy": "strict-origin-when-cross-origin",
    }

    def __init__(self, logger):
        super().__init__(logger)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        attr_config = {
            "interval": {
                "default": 3,
                "format": "int",
                "help": "Change interval in hours, default is 3 hours.",
            }
        }
        await self.set_attr_config(attr_config)

        if len(args) != 4:
            return "Usage: define gh.pixel fhempy geizhals URL"

        self.url = args[3]
        self.product_id = self.url[self.url.rfind("-a") + 2 : self.url.rfind(".html")]
        self.location = self.url[
            self.url.find("geizhals") + 9 : self.url.find("geizhals") + 11
        ]
        await fhem.readingsSingleUpdateIfChanged(
            self.hash,
            "link",
            '<html><a href="'
            + self.url
            + '" target="_blank">Open geizhals '
            + "(new window/tab)</a><br></html>",
            1,
        )

        self.create_async_task(self.update_loop())

    async def update_product_page_infos(self):
        async with aiohttp.ClientSession() as session:
            async with session.get(
                self.url,
                headers=geizhals.headers,
            ) as resp:
                if resp.status == 200:
                    html = await resp.text()
                    await utils.run_blocking(
                        functools.partial(self.handle_product_page, html)
                    )

    def handle_product_page(self, html):
        soup = BeautifulSoup(html, "html.parser")
        self.product_name = soup.find("h1", {"class": "variant__header__headline"}).text
        self.store = soup.find("span", {"class": "notrans"}).text
        self.store_availability = (
            soup.find("div", {"id": "offer__0"})
            .find("div", {"class": "offer__delivery-time"})
            .text
        )

    async def update_loop(self):
        # get product page infos
        await self.update_product_page_infos()
        if await fhem.AttrVal(self.hash["NAME"], "alias", "") == "":
            await fhem.CommandAttr(
                self.hash, f"{self.hash['NAME']} alias {self.product_name}"
            )

        while True:
            # aiohttp post
            data = {"id": self.product_id, "params": {"days": 31, "loc": self.location}}
            try:
                await fhem.readingsBeginUpdate(self.hash)
                async with aiohttp.ClientSession() as session:
                    async with session.post(
                        "https://geizhals.at/api/gh0/price_history",
                        headers=geizhals.headers,
                        json=data,
                    ) as resp:
                        if resp.status == 200:
                            await self.handle_response(await resp.json())
                        else:
                            await fhem.readingsSingleUpdate(
                                self.hash,
                                "state",
                                f"failed: HTTP error {resp.status}",
                                1,
                            )
                            self.logger.error(
                                f"Failed to fetch from geizhals, "
                                f"failed with status {resp.status}"
                            )

                await self.update_product_page_infos()
                await fhem.readingsBulkUpdateIfChanged(self.hash, "store", self.store)
                await fhem.readingsBulkUpdateIfChanged(
                    self.hash, "store_availability", self.store_availability
                )

            except Exception:
                self.logger.exception("Failed to update")

            await fhem.readingsEndUpdate(self.hash, 1)
            await asyncio.sleep(self._attr_interval * 60 * 60)

    async def handle_response(self, response):
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "last_update", response["meta"]["last_formatted"]
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "current_price_euro", response["meta"]["current_best"]
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "max", response["meta"]["max"]
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "min", response["meta"]["min"]
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash,
            "state",
            str(response["meta"]["current_best"])
            + " € ("
            + str(response["meta"]["min"])
            + " € - "
            + str(response["meta"]["max"])
            + " €)",
        )



"""
Support for eq3 Bluetooth Smart thermostats.

All temperatures in Celsius.

To get the current state, update() has to be called for powersaving reasons.
Schedule needs to be requested with query_schedule() before accessing for similar reasons.
"""

# handle: 0x0110, char properties: 0x02, char value handle: 0x0111, uuid: 00002a00-0000-1000-8000-00805f9b34fb
# handle: 0x0120, char properties: 0x02, char value handle: 0x0121, uuid: 00002a01-0000-1000-8000-00805f9b34fb
# handle: 0x0130, char properties: 0x02, char value handle: 0x0131, uuid: 00002a02-0000-1000-8000-00805f9b34fb
# handle: 0x0140, char properties: 0x08, char value handle: 0x0141, uuid: 00002a03-0000-1000-8000-00805f9b34fb
# handle: 0x0150, char properties: 0x02, char value handle: 0x0151, uuid: 00002a04-0000-1000-8000-00805f9b34fb
# handle: 0x0210, char properties: 0x22, char value handle: 0x0211, uuid: 00002a05-0000-1000-8000-00805f9b34fb
# handle: 0x0310, char properties: 0x02, char value handle: 0x0311, uuid: 00002a29-0000-1000-8000-00805f9b34fb
# handle: 0x0320, char properties: 0x02, char value handle: 0x0321, uuid: 00002a24-0000-1000-8000-00805f9b34fb
# handle: 0x0410, char properties: 0x0a, char value handle: 0x0411, uuid: 3fa4585a-ce4a-3bad-db4b-b8df8179ea09
# handle: 0x0420, char properties: 0x1a, char value handle: 0x0421, uuid: d0e8434d-cd29-0996-af41-6c90f4e0eb2a
# handle: 0xff01, char properties: 0x38, char value handle: 0xff02, uuid: e3dd50bf-f7a7-4e99-838e-570a086c666b
# handle: 0xff04, char properties: 0x08, char value handle: 0xff05, uuid: 92e86c7a-d961-4091-b74f-2409e72efe36
# handle: 0xff06, char properties: 0x02, char value handle: 0xff07, uuid: 347f7608-2e2d-47eb-913b-75d4edc4de3b

import asyncio
import codecs
import struct
from datetime import datetime, timedelta
from enum import IntEnum

from construct import Byte

from ..core import bluetoothle
from .structures import AwayDataAdapter, DeviceId, Schedule, Status

PROP_WRITE_HANDLE = "3fa4585a-ce4a-3bad-db4b-b8df8179ea09"
PROP_NTFY_HANDLE = "d0e8434d-cd29-0996-af41-6c90f4e0eb2a"

PROP_ID_QUERY = 0
PROP_ID_RETURN = 1
PROP_INFO_QUERY = 3
PROP_INFO_RETURN = 2
PROP_COMFORT_ECO_CONFIG = 0x11
PROP_OFFSET = 0x13
PROP_WINDOW_OPEN_CONFIG = 0x14
PROP_SCHEDULE_QUERY = 0x20
PROP_SCHEDULE_RETURN = 0x21

PROP_MODE_WRITE = 0x40
PROP_TEMPERATURE_WRITE = 0x41
PROP_COMFORT = 0x43
PROP_ECO = 0x44
PROP_BOOST = 0x45
PROP_LOCK = 0x80

EQ3BT_AWAY_TEMP = 12.0
EQ3BT_MIN_TEMP = 5.0
EQ3BT_MAX_TEMP = 29.5
EQ3BT_OFF_TEMP = 4.5
EQ3BT_ON_TEMP = 30.0


class Mode(IntEnum):
    """Thermostat modes."""

    Unknown = -1
    Closed = 0
    Open = 1
    Auto = 2
    Manual = 3
    Away = 4
    Boost = 5


MODE_NOT_TEMP = [Mode.Unknown, Mode.Closed, Mode.Open]


class TemperatureException(Exception):
    """Temperature out of range error."""

    pass


# pylint: disable=too-many-instance-attributes
class Thermostat:
    """Representation of a EQ3 Bluetooth Smart thermostat."""

    def __init__(
        self, logger, hash, _mac, keep_connection=True, notification_callback=None
    ):
        """Initialize the thermostat."""
        self.logger = logger

        self._target_temperature = 0
        self._mode = Mode.Unknown
        self._valve_state = 0
        self._raw_mode = None

        self._schedule = {}

        self._window_open_temperature = None
        self._window_open_time = None
        self._comfort_temperature = None
        self._eco_temperature = None
        self._temperature_offset = None

        self._away_temp = EQ3BT_AWAY_TEMP
        self._away_duration = timedelta(days=30)
        self._away_end = None

        self._firmware_version = None
        self._device_serial = None

        self._notification_callback = notification_callback

        self._conn = bluetoothle.BluetoothLE(
            self.logger,
            hash,
            _mac,
        )
        self._conn.register_notification_listener(self.handle_notification)

    async def connect(self):
        await self._conn.connect()

    def __str__(self):
        away_end = "no"
        if self.away_end:
            away_end = "end: %s" % self._away_end

        return "[%s] Target %s (mode: %s, away: %s)" % (
            self._conn.mac,
            self.target_temperature,
            self.mode_readable,
            away_end,
        )

    def _verify_temperature(self, temp):
        """Verifies that the temperature is valid.
        :raises TemperatureException: On invalid temperature.
        """
        if temp < self.min_temp or temp > self.max_temp:
            raise TemperatureException(
                "Temperature {} out of range [{}, {}]".format(
                    temp, self.min_temp, self.max_temp
                )
            )

    def parse_schedule(self, data):
        """Parses the device sent schedule."""
        sched = Schedule.parse(data)
        self.logger.debug("Got schedule data for day '%s'", sched.day)

        return sched

    async def disconnect(self):
        await self._conn.disconnect()

    def handle_notification(self, uuid, data):
        """Handle Callback from a Bluetooth (GATT) request."""
        self.logger.debug("Received notification from the device..")

        if data[0] == PROP_INFO_RETURN and data[1] == 1:
            self.logger.debug("Got status: %s" % codecs.encode(data, "hex"))
            status = Status.parse(data)
            self.logger.debug("Parsed status: %s", status)

            self._raw_mode = status.mode
            self._valve_state = status.valve
            self._target_temperature = status.target_temp

            if status.mode.BOOST:
                self._mode = Mode.Boost
            elif status.mode.AWAY:
                self._mode = Mode.Away
                self._away_end = status.away
            elif status.mode.MANUAL:
                if status.target_temp == EQ3BT_OFF_TEMP:
                    self._mode = Mode.Closed
                elif status.target_temp == EQ3BT_ON_TEMP:
                    self._mode = Mode.Open
                else:
                    self._mode = Mode.Manual
            else:
                self._mode = Mode.Auto

            presets = status.presets
            if presets:
                self._window_open_temperature = presets.window_open_temp
                self._window_open_time = presets.window_open_time
                self._comfort_temperature = presets.comfort_temp
                self._eco_temperature = presets.eco_temp
                self._temperature_offset = presets.offset
            else:
                self._window_open_temperature = None
                self._window_open_time = None
                self._comfort_temperature = None
                self._eco_temperature = None
                self._temperature_offset = None

            self.logger.debug("Valve state:      %s", self._valve_state)
            self.logger.debug("Mode:             %s", self.mode_readable)
            self.logger.debug("Target temp:      %s", self._target_temperature)
            self.logger.debug("Away end:         %s", self._away_end)
            self.logger.debug("Window open temp: %s", self._window_open_temperature)
            self.logger.debug("Window open time: %s", self._window_open_time)
            self.logger.debug("Comfort temp:     %s", self._comfort_temperature)
            self.logger.debug("Eco temp:         %s", self._eco_temperature)
            self.logger.debug("Temp offset:      %s", self._temperature_offset)

        elif data[0] == PROP_SCHEDULE_RETURN:
            parsed = self.parse_schedule(data)
            self._schedule[parsed.day] = parsed

        elif data[0] == PROP_ID_RETURN:
            parsed = DeviceId.parse(data)
            self.logger.debug("Parsed device data: %s", parsed)
            self._firmware_version = parsed.version
            self._device_serial = parsed.serial

        else:
            self.logger.debug(
                "Unknown notification %s (%s)", data[0], codecs.encode(data, "hex")
            )
        if self._notification_callback:
            asyncio.create_task(self._notification_callback())

    async def query_id(self):
        """Query device identification information, e.g. the serial number."""
        self.logger.debug("Querying id..")
        value = struct.pack("B", PROP_ID_QUERY)
        await self._conn.write_gatt_char(PROP_WRITE_HANDLE, value)

    async def update(self):
        """Update the data from the thermostat. Always sets the current time."""
        self.logger.debug("Querying the device..")
        time = datetime.now()
        value = struct.pack(
            "BBBBBBB",
            PROP_INFO_QUERY,
            time.year % 100,
            time.month,
            time.day,
            time.hour,
            time.minute,
            time.second,
        )

        await self._conn.write_gatt_char(PROP_WRITE_HANDLE, value)

    async def query_schedule(self, day):
        self.logger.debug("Querying schedule..")

        if day < 0 or day > 6:
            self.logger.error("Invalid day: %s", day)

        value = struct.pack("BB", PROP_SCHEDULE_QUERY, day)

        await self._conn.write_gatt_char(PROP_WRITE_HANDLE, value)

    @property
    def schedule(self):
        """Returns previously fetched schedule.
        :return: Schedule structure or None if not fetched.
        """
        return self._schedule

    async def set_schedule(self, data):
        """Sets the schedule for the given day."""
        value = Schedule.build(data)
        await self._conn.write_gatt_char(PROP_WRITE_HANDLE, value)

    @property
    def target_temperature(self):
        """Return the temperature we try to reach."""
        return self._target_temperature

    async def set_target_temperature(self, temperature):
        """Set new target temperature."""
        dev_temp = int(temperature * 2)
        if temperature == EQ3BT_OFF_TEMP or temperature == EQ3BT_ON_TEMP:
            dev_temp |= 0x40
            value = struct.pack("BB", PROP_MODE_WRITE, dev_temp)
        else:
            self._verify_temperature(temperature)
            value = struct.pack("BB", PROP_TEMPERATURE_WRITE, dev_temp)

        await self._conn.write_gatt_char(PROP_WRITE_HANDLE, value)

    @property
    def mode(self):
        """Return the current operation mode"""
        return self._mode

    async def set_mode(self, mode):
        """Set the operation mode."""
        self.logger.debug("Setting new mode: %s", mode)

        if self.mode == Mode.Boost and mode != Mode.Boost:
            self.boost = False

        if mode == Mode.Boost:
            self.boost = True
            return
        elif mode == Mode.Away:
            end = datetime.now() + self._away_duration
            return self.set_away(end, self._away_temp)
        elif mode == Mode.Closed:
            return await self.send_mode(0x40 | int(EQ3BT_OFF_TEMP * 2))
        elif mode == Mode.Open:
            return await self.send_mode(0x40 | int(EQ3BT_ON_TEMP * 2))

        if mode == Mode.Manual:
            temperature = max(
                min(self._target_temperature, self.max_temp), self.min_temp
            )
            return await self.send_mode(0x40 | int(temperature * 2))
        else:
            return await self.send_mode(0)

    @property
    def away_end(self):
        return self._away_end

    async def set_away(self, away_end=None, temperature=EQ3BT_AWAY_TEMP):
        """Sets away mode with target temperature.
        When called without parameters disables away mode."""
        if not away_end:
            self.logger.debug("Disabling away, going to auto mode.")
            return await self.send_mode(0x00)

        self.logger.debug("Setting away until %s, temp %s", away_end, temperature)
        adapter = AwayDataAdapter(Byte[4])
        packed = adapter.build(away_end)

        await self.send_mode(0x80 | int(temperature * 2), packed)

    async def send_mode(self, mode, payload=None):
        value = struct.pack("BB", PROP_MODE_WRITE, mode)
        if payload:
            value += payload
        await self._conn.write_gatt_char(PROP_WRITE_HANDLE, value)

    @property
    def mode_readable(self):
        """Return a readable representation of the mode.."""
        ret = ""
        mode = self._raw_mode
        if mode is None:
            return "unknown"

        if mode.MANUAL:
            ret = "manual"
            if self.target_temperature < self.min_temp:
                ret += " off"
            elif self.target_temperature >= self.max_temp:
                ret += " on"
            else:
                ret += " (%sC)" % self.target_temperature
        else:
            ret = "auto"

        if mode.AWAY:
            ret += " holiday"
        if mode.BOOST:
            ret += " boost"
        if mode.DST:
            ret += " dst"
        if mode.WINDOW:
            ret += " window"
        if mode.LOCKED:
            ret += " locked"
        if mode.LOW_BATTERY:
            ret += " low battery"

        return ret

    @property
    def boost(self):
        """Returns True if the thermostat is in boost mode."""
        return self.mode == Mode.Boost

    async def set_boost(self, boost):
        """Sets boost mode."""
        self.logger.debug("Setting boost mode: %s", boost)
        value = struct.pack("BB", PROP_BOOST, bool(boost))
        await self._conn.write_gatt_char(PROP_WRITE_HANDLE, value)

    @property
    def valve_state(self):
        """Returns the valve state. Probably reported as percent open."""
        return self._valve_state

    @property
    def window_open(self):
        """Returns True if the thermostat reports a open window
        (detected by sudden drop of temperature)"""
        return self._raw_mode and self._raw_mode.WINDOW

    async def window_open_config(self, temperature, duration):
        """Configures the window open behavior. The duration is specified in
        5 minute increments."""
        self.logger.debug(
            "Window open config, temperature: %s duration: %s", temperature, duration
        )
        self._verify_temperature(temperature)
        if duration < 0 and duration > 3600:
            raise ValueError

        value = struct.pack(
            "BBB",
            PROP_WINDOW_OPEN_CONFIG,
            int(temperature * 2),
            int(duration / 300),
        )
        await self._conn.write_gatt_char(PROP_WRITE_HANDLE, value)

    @property
    def window_open_temperature(self):
        """The temperature to set when an open window is detected."""
        return self._window_open_temperature

    @property
    def window_open_time(self):
        """Timeout to reset the thermostat after an open window is detected."""
        return self._window_open_time

    @property
    def locked(self):
        """Returns True if the thermostat is locked."""
        return self._raw_mode and self._raw_mode.LOCKED

    async def set_locked(self, lock):
        """Locks or unlocks the thermostat."""
        self.logger.debug("Setting the lock: %s", lock)
        value = struct.pack("BB", PROP_LOCK, bool(lock))
        await self._conn.write_gatt_char(PROP_WRITE_HANDLE, value)

    @property
    def low_battery(self):
        """Returns True if the thermostat reports a low battery."""
        return self._raw_mode and self._raw_mode.LOW_BATTERY

    async def temperature_presets(self, comfort, eco):
        """Set the thermostats preset temperatures comfort (sun) and
        eco (moon)."""
        self.logger.debug(
            "Setting temperature presets, comfort: %s eco: %s", comfort, eco
        )
        self._verify_temperature(comfort)
        self._verify_temperature(eco)
        value = struct.pack(
            "BBB", PROP_COMFORT_ECO_CONFIG, int(comfort * 2), int(eco * 2)
        )
        await self._conn.write_gatt_char(PROP_WRITE_HANDLE, value)

    @property
    def comfort_temperature(self):
        """Returns the comfort temperature preset of the thermostat."""
        return self._comfort_temperature

    @property
    def eco_temperature(self):
        """Returns the eco temperature preset of the thermostat."""
        return self._eco_temperature

    @property
    def temperature_offset(self):
        """Returns the thermostat's temperature offset."""
        return self._temperature_offset

    async def set_temperature_offset(self, offset):
        """Sets the thermostat's temperature offset."""
        self.logger.debug("Setting offset: %s", offset)
        # [-3,5 .. 0  .. 3,5 ]
        # [00   .. 07 .. 0e ]
        if offset < -3.5 or offset > 3.5:
            raise TemperatureException("Invalid value: %s" % offset)

        current = -3.5
        values = {}
        for i in range(15):
            values[current] = i
            current += 0.5

        value = struct.pack("BB", PROP_OFFSET, values[offset])
        await self._conn.write_gatt_char(PROP_WRITE_HANDLE, value)

    async def activate_comfort(self):
        """Activates the comfort temperature."""
        value = struct.pack("B", PROP_COMFORT)
        await self._conn.write_gatt_char(PROP_WRITE_HANDLE, value)

    async def activate_eco(self):
        """Activates the comfort temperature."""
        value = struct.pack("B", PROP_ECO)
        await self._conn.write_gatt_char(PROP_WRITE_HANDLE, value)

    @property
    def min_temp(self):
        """Return the minimum temperature."""
        return EQ3BT_MIN_TEMP

    @property
    def max_temp(self):
        """Return the maximum temperature."""
        return EQ3BT_MAX_TEMP

    @property
    def firmware_version(self):
        """Return the firmware version."""
        return self._firmware_version

    @property
    def device_serial(self):
        """Return the device serial number."""
        return self._device_serial



import asyncio

import aiohttp
from bs4 import BeautifulSoup

from .. import fhem, generic


class volvo_software_update(generic.FhemModule):
    def __init__(self, logger):
        super().__init__(logger)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        
        attr_config = {
            "interval": {
                "default": 1800,
                "format": "int",
                "help": "Change interval, default is 1800s.",
            }
        }
        await self.set_attr_config(attr_config)

        if len(args) != 4:
            return "Usage: define my_volvo_update fhempy volvo_software_update URL"

        self.update_url = args[3]

        self.create_async_task(self.update_loop())

    async def update_loop(self):
        while True:
            # aiohttp get
            headers = {
                "accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9",
                "accept-language": "en-AT,en;q=0.9,de-AT;q=0.8,de;q=0.7,en-GB;q=0.6,en-US;q=0.5",
                "sec-ch-ua": '"Not?A_Brand";v="8", "Chromium";v="108", "Google Chrome";v="108"',
                "sec-ch-ua-mobile": "?0",
                "sec-ch-ua-platform": '"Chrome OS"',
                "sec-fetch-dest": "document",
                "sec-fetch-mode": "navigate",
                "sec-fetch-site": "none",
                "sec-fetch-user": "?1",
                "upgrade-insecure-requests": "1",
                "User-Agent": "Mozilla/5.0 (X11; CrOS x86_64 14541.0.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/108.0.0.0 Safari/537.36",
            }
            try:
                async with aiohttp.ClientSession(headers=headers) as session:
                    async with session.get(self.update_url) as resp:
                        if resp.status == 200:
                            await self.handle_response(await resp.text())
                        else:
                            await fhem.readingsSingleUpdate(
                                self.hash,
                                "state",
                                f"failed: HTTP error {resp.status}",
                                1,
                            )
                            self.logger.error(
                                f"Failed to fetch {self.update_url}, "
                                f"failed with status {resp.status}"
                            )
            except Exception:
                self.logger.exception("Failed to update")
            await asyncio.sleep(self._attr_interval)

    async def handle_response(self, response):
        # bs4
        soup = BeautifulSoup(response, "html.parser")
        entries = soup.findAll("div", {"class": "segment"})
        for entry in entries:
            if entry.find("h2"):
                await fhem.readingsBeginUpdate(self.hash)
                await fhem.readingsBulkUpdateIfChanged(
                    self.hash, "latest_release_notes", f"<html>{entry}</html>"
                )
                release_text = entry.find("h2").text

                if release_text.find(" V") > 0:
                    release_number = release_text[release_text.find(" V") + 2 :]
                    state_text = "Version "
                else:
                    release_number = release_text
                    state_text = ""

                await fhem.readingsBulkUpdateIfChanged(
                    self.hash, "latest_release", release_number
                )
                await fhem.readingsBulkUpdateIfChanged(
                    self.hash, "state", f"{state_text}{release_number}"
                )
                latest_update = soup.findAll("em")[-1].text
                await fhem.readingsBulkUpdateIfChanged(
                    self.hash, "latest_update", latest_update
                )
                await fhem.readingsEndUpdate(self.hash, 1)
                break
